import sys
from .version import __version__
from .main import exosomes_recognizer, cellfree_simulator
from .functional import source_tracker, ESAI_celltype


__all__ = [
    "exosomes_recognizer",
    "cellfree_simulator",
    "source_tracker",
    "ESAI_celltype",
]



from .utils import *
import os
import sys

def exoTras_command():
    ''' Example of taking inputs for exoTras'''
    args = sys.argv[1:]
    if len(args) < 1:
        print("usage: exosomes_recognizer, source_tracker, ESAI_celltype, cellfree_simulator")


def cellfree_simulator(out_path, gene_exp_exo, gene_exp_cell, expect_UMI = [40, 70, 100, 130], exosomes_fraction = [0.005, 0.01, 0.05, 0.10], exosomes=500):
    all_num = [exosomes/f for f in exosomes_fraction]
    for i in expect_UMI:
        for j in all_num:
            exo_count = np.random.poisson(lam=i, size=exosomes)
            all_count = np.random.poisson(lam=i, size=j)

            exo_data = process_random(gene_exp_exo, exo_count)
            exo_data.columns = ['exo'+str(i) for i in exo_data.columns.values]
            cell_data = process_random(gene_exp_cell, all_count)
            exo_cell_data = pd.concat([cell_data, exo_data], axis=1).fillna(0)

            new_index = [ i.split(' /// ')[-1].split('-')[0] for i in exo_cell_data.index.values]
            exo_cell_data.index = new_index
            exo_cell_data.to_csv(out_path + '/simulation_UMI' + str(i) + '_fraction' + str(int(exosomes/j)) + '.csv')
    
    return('Simulation done!')

def exosomes_aggregator(out_path, name_list, score_t, threads, flag=0):
    if flag == 0:
        adata_com = read_project(out_path, name_list)

        genes_out = get_genes(out_path, name_list)
        genes = count_genes(genes_out, out_path)
        iteration_list = set(genes['0'])#exo_list_s
        inter_gene = []
        inter_out_p = pd.DataFrame(list(range(0, adata_com.obs.shape[0])))
        pth = 0.005
        same = [i for i in iteration_list if i in adata_com.var_names]
        thershold = 0
        max_M = max(adata_com.obs['n_genes'])

        adata_com = final_menrich(adata_com, same, thershold, max_M, threads)
        # adata_com.write('./all.h5ad')

        adata_com.obs['exoTras'] = ['Exosomes' if i else 'False' for i in (adata_com.obs['total_counts'] < 200) & (adata_com.obs['score'] < score_t)]

        ## write files in this project
        adata_exo = copy.copy(adata_com[(adata_com.obs['total_counts'] < 200) & (adata_com.obs['score'] < score_t), :]) 
        adata_exo = adata_exo.copy()
        adata_com.obs['score'] = -np.log10(adata_com.obs['score'] + adata_com.obs.loc[(adata_com.obs['score']>0).values, 'score'].min()/10)
        adata_exo.obs['score'] = -np.log10(adata_exo.obs['score'] + adata_exo.obs.loc[(adata_exo.obs['score']>0).values, 'score'].min()/10)
        adata_com.write(str(out_path) + '/raw_' + 'exoTras' + '.h5ad')
        adata_exo.write(str(out_path) + '/exosomes_exoTras.h5ad')
    
    return('Aggregation done!')


def exosomes_recognizer(sample_file, out_path, input_path=None, species='Homo', predefine_threads=-2, get_only=False, score_t = None):
    if predefine_threads == -2:
        threads = cpu_count()-2
    else:
        threads = predefine_threads
    
    if out_path.endswith('/'):
        out_path = out_path[:-1]
    
    if score_t is not None:
        exo_list, score_t = get_exo_list(species, score_t)
    else:
        exo_list, score_t = get_exo_list(species)
    score_t = float(score_t)
    exo_list_s = list(set(exo_list))
    
    sample_log = get_sample(sample_file)
    len_sample = len(sample_log)
    name_list = []

    for sample in sample_log:
        ##read adata
        if input_path!= None:
            each_adata = str(input_path) + '/' + str(sample)
        else:
            each_adata = sample

        adata = read_adata(each_adata, get_only=False)
        adata = filter_adata(adata)
        sample = sample.replace(".", "_")
        name_list.append(sample)

        iteration_list = exo_list_s
        inter_gene = []
        pth = 0.005
        flag = 1
        inter_out_p = pd.DataFrame(list(range(0, adata.obs.shape[0])))
        inter_adata = copy.copy(adata[(adata.obs['total_counts'] < 200) & (adata.obs['total_counts'] > 20),:])
        inter_adata = inter_adata.copy()
        sc.pp.normalize_total(inter_adata, target_sum=1e2)
        max_M = max(adata.obs['n_genes'])
        # inter_adata.X = (inter_adata.X / inter_adata.X.sum(1).reshape(inter_adata.shape[0], 1))*100
        ##iterations 
        for itera in range(10):
            if(len(iteration_list) == 0):
                print('no genes')
                flag = 0
                break

            thershold = 0
            iteration_list_old = iteration_list
            inter_adata, iteration_list = iteration(inter_adata, iteration_list, thershold, max_M, threads=threads, number_g = 30, alpha = 0.10)
            print(itera, len(iteration_list))
                
            if(len(set(iteration_list_old) & set(iteration_list)) == max(len(iteration_list_old), len(iteration_list))):
                break
            else:
                inter_gene.append(iteration_list_old)
                inter_out_p = pd.concat((inter_out_p, inter_adata.obs['score']), axis=1, ignore_index=True)

        if flag == 1:
            genes = iteration_list[0:15]
            same = [i for i in (genes) if i in adata.var_names]
            thershold = 0
            adata = final_menrich(adata, same, thershold, max_M, threads)

        elif (len(iteration_list_old) > 0) & (len(iteration_list_old) < len(exo_list_s)):
            genes = iteration_list_old[0:15]
            same = [i for i in (genes) if i in adata.var_names]
            thershold = 0
            adata = final_menrich(adata, same, thershold, max_M, threads)

        else:
            print('Exosomes undetectable  in ' + str(sample))
            continue
        
        ## write files
        if len_sample > 1:
            if os.path.isdir(str(out_path) + '/tmp_out'):
                pass
            else:
                os.mkdir(str(out_path) + '/tmp_out')

            if os.path.isdir(str(out_path) + '/tmp_out/' + sample):
                pass
            else:
                os.mkdir(str(out_path) + '/tmp_out/' + sample)

            adata.write(str(out_path) + '/tmp_out/' + sample + '/raw_' + sample + '.h5ad')
            with open(str(out_path) +'/tmp_out/' + sample + '/itera_gene.txt', 'wb') as fp:
                pickle.dump(inter_gene, fp)
        else:
            adata.obs['exoTras'] = ['Exosomes' if i else 'False' for i in (adata.obs['total_counts'] < 200) & (adata.obs['score'] < score_t)]
            adata.obs['score'] = -np.log10(adata.obs['score'] + adata.obs.loc[(adata.obs['score']>0).values, 'score'].min()/10)
            adata.write(str(out_path) + '/raw_' + str(sample) + '.h5ad')
            with open(str(out_path)+ '/' + str(sample) + "_exo.txt", "w") as f:
                for s in genes:
                    f.write(str(s) +"\n")

    ## Aggregation
    if len_sample > 1:
        exosomes_aggregator(out_path, name_list, score_t, threads)

    return('Recognization done!')



import numpy as np
import pandas as pd
import scanpy as sc
import scipy
import copy
import sys
import os
import pickle
import statsmodels.api as sm
from multiprocessing import Pool, cpu_count


def get_sample(sample_log):
    names_list = []
    with open(sample_log, 'r') as  f:
        for line in f.readlines():
            line = line.strip()
            names_list.append(line)
    
    return(names_list)


def read_adata(adata_path, get_only=False):
    if (adata_path).endswith('.h5'):
        adata = sc.read_10x_h5(adata_path)
    elif (adata_path).endswith('.h5ad'):
        adata = sc.read_h5ad(adata_path)
    else:
        adata = sc.read_10x_mtx(adata_path + '/outs/raw_feature_bc_matrix/')

    return(adata)

import anndata
def read_project(output_path, names_list):

    names_list_1 = copy.copy(names_list)
    adata_list = []
    for i in (names_list):
        i = i.replace(".", "_")
        if os.path.exists(output_path + '/tmp_out/' + i):
            
            # exec('g{} = sc.read_h5ad("{}/{}/{}/raw_{}.h5ad")'.format(i, str(output_path), 'tmp_out', i, i), d1)
            adata_list.append(sc.read_h5ad(str(output_path) + '/tmp_out/'+ str(i) + '/raw_' + str(i) + '.h5ad'))
            adata_list[-1].var_names_make_unique()
            #d1['g'+i].var_names_make_unique()
            # exec('g{}.var_names_make_unique()'.format(i), globals())
            # exec('adata_list.append(g{})'.format(i))
        else:
            print(i)
            names_list_1.remove(i)
    #d = {}
    #exec('adata_com = adata_list[0].concatenate(d1[g{}], batch_categories = names_list_1)'.format(names_list_1[0], "], d1[g".join(names_list_1[1:])), d)
    #adata_com = d['adata_com']
    #adata_com.write(output_path+'/all.h5ad')
    adata_com = anndata.concat(adata_list, label="batch", keys=names_list_1)#adata_list[0].concatenate(adata_list[1:], batch_categories = names_list_1)
    adata_com.obs_names_make_unique()

    return(adata_com)

def get_genes(output_path, names_list):
    genes_out = []
    for i in names_list:
        path_gene = str(output_path) + '/tmp_out/' + i + '/itera_gene.txt'
        if os.path.exists(path_gene):
            with open (path_gene, 'rb') as fp:
                inter_gene = pickle.load(fp)
                genes_out.append(inter_gene[len(inter_gene)-1])
    
    return(genes_out)

def count_genes(genes_out, out_path):
    cal_gene_out = []
    for i in genes_out:
        cal_gene_out.extend(list(i))

    pd.Series(cal_gene_out).value_counts().head(15).index.to_series().to_csv(str(out_path) + '/exo_genes.csv')
    genes = pd.read_csv(str(out_path) + '/exo_genes.csv')

    return(genes)


def filter_adata(adata):
    adata.var_names_make_unique()
    adata.var['mt'] = adata.var_names.str.startswith(('mt-', 'MT-'))
    adata.var['ribo'] = adata.var_names.str.startswith(('Rps', 'Rpl', 'RPS', 'RPL'))
    sc.pp.calculate_qc_metrics(adata, qc_vars=['mt', 'ribo'], inplace=True, percent_top='', log1p =False)
    adata = adata[(adata.obs.pct_counts_mt < 15) & (adata.obs.pct_counts_ribo < 30), :]
    sc.pp.filter_cells(adata, min_genes=6)
    sc.pp.filter_genes(adata, min_cells=5)
    if 'n_genes_by_counts' in adata.obs.columns:
        adata.obs['n_genes'] = adata.obs['n_genes_by_counts']

    return(adata)

from pathlib import Path
def get_exo_list(species, score_t='1e-15'):
    ## get initial gene set
    ##exo gene symbols
    if species == 'Homo':
        exo_list = [x.strip() for x in open(Path(__file__).parent / 'id_homo_rna_exo')]
    elif species == 'Mus':
        exo_list = [x.strip() for x in open(Path(__file__).parent / 'id_mus_rna_exo')]
        score_t = '1e-10'
    
    exo_list_count = pd.value_counts(exo_list, sort=False)
    exo_list = exo_list_count[exo_list_count > 0].index
    
    return([exo_list, score_t])

from scipy.stats import hypergeom
from exoTras import env
def process_inter(i):
    k = (env.Inter_adata[i, env.Same].X > env.Thershold).sum()
    M = env.Max_M
    n = len(env.Same)
    N = int(env.Inter_adata[i, :].obs['n_genes'])
    return([hypergeom.sf(k, M, n, N), N, k/N, n/M])

def corr_genes(i):
    #i, inter_adata
    if (len(np.unique(env.Inter_adata.X.A[:, i])) > 1):
        tmp = scipy.stats.spearmanr(env.Inter_adata.X.A[:, i], -np.log(env.Inter_adata.obs['score'] + 10**-50))
        return(tmp.correlation)
    else:
        return(-1)

def multi_enrich(inter_adata, same, iteration_list, thershold, max_M, threads):
    len((iteration_list)), len(same), max(inter_adata.obs['n_genes'])

    out0 = []
    item_list = range(inter_adata.n_obs)
    pool = Pool(threads, env.initializer, (inter_adata, same, thershold, max_M))
    out0 = pool.map(process_inter, item_list)
    pool.close()
    pool.join()

    ##
    out = pd.DataFrame(out0)
    # out.to_csv('./out.csv')
    out['score'] = sm.stats.fdrcorrection(out[0])[1]
    # iteration_list_old = iteration_list
    
    inter_adata.obs['score'] = out['score'].tolist()

    return(inter_adata)


def multi_cor(inter_adata, threads):
    result_tmp = []

    pool = Pool(threads, env.initializer_simple, (inter_adata,))
    item_list = range(inter_adata.X.shape[1])
    
    result_tmp = pool.map(corr_genes, item_list)
    pool.close()
    pool.join()

    return(result_tmp)

def representative_gene(result_tmp, number_g = 30, alpha = 0.10):
    max_number = []
    max_index = []
    t = copy.deepcopy(result_tmp)
    for _ in range(number_g):
        number = max(t)
        index = t.index(number)
        t[index] = -1
        if (number > alpha):
            max_index.append(index)
            max_number.append(number)

    return([max_index, max_number])

def get_iteration(inter_adata, max_index):
    # p_thre_filter = min(max_number)
    #tt_list  = inter_adata.var.index[[(i >= 0) & (i >= p_thre_filter) for i in result_tmp]]
    tt_list = inter_adata.var.index[max_index]
    iteration_list = list(tt_list)# & exo_list_s  set()

    return(iteration_list)

def iteration(inter_adata, iteration_list, thershold, max_M, threads=20, number_g = 30, alpha = 0.10):
    same = [i for i in iteration_list if i in inter_adata.var_names]
    inter_adata = multi_enrich(inter_adata, same, iteration_list, thershold, max_M, threads)
    result_tmp = multi_cor(inter_adata, threads)

    
    max_index, max_number = representative_gene(result_tmp, number_g = 30, alpha = 0.10)
    iteration_list = get_iteration(inter_adata, max_index)

    return(inter_adata, iteration_list)
    
def process(i):
    k = (env.Adata[i, env.Same].X > env.Thershold).sum()
    M = env.Max_M
    n = len(env.Same)
    N = int(env.Adata[i, :].obs['n_genes'])
    return([scipy.stats.hypergeom.sf(k, M, n, N), N, k/N, n/M])


def final_menrich(adata, same, thershold, max_M, threads):

    out0 = []
    item_list = range(adata.n_obs)
    pool = Pool(threads, env.initializer_adata, (adata, same, thershold, max_M))
    out0 = pool.map(process, item_list)
    pool.close()
    pool.join()

    ##
    out = pd.DataFrame(out0)
    # out.to_csv('./out.csv')
    out['score'] = sm.stats.fdrcorrection(out[0])[1]
    #iteration_list_old = iteration_list

    adata.obs['score'] = out['score'].tolist()
    
    return(adata)



## cell free droplets simulation
def zinb_genes(i, gene_exp_a, num):
    cells = num #number of cells
    n = 1
    p = 0.8# 1 - droplet rate
    
    exp = gene_exp_a.iloc[i]#exp_i# gene expression mean
    success = 1
    
    out_zinb = np.random.negative_binomial(success, success/(success+exp), size=cells) * \
        np.random.binomial(n, p, size=cells)
    return(pd.Series(data = out_zinb, name = gene_exp_a.index[i]))
    
def run_weigths(gene_exp_a, num=1000):
    
    weight = []
    def log_result(result):
        # This is called whenever foo_pool(i) returns a result.
        # result_list is modified only by the main process, not the pool workers.
        weight.append(result)

    pool = Pool(cpu_count()-4)

    for i in range(len(gene_exp_a)):
        pool.apply_async(zinb_genes, args = (i, gene_exp_a, num), callback = log_result)
        
    pool.close()
    pool.join()
    
#     weight = (gene_exp_a.apply(zinb_genes, axis = 1))
    
    weight = pd.DataFrame(weight)#.T.fillna(0)
    
    return(weight)


def process_random(gene_exp, count):
    num = len(count)
    out = run_weigths(gene_exp, num)
    
    return(round((out / out.sum(0)) * count))









Adata = None
Inter_adata = None
Thershold = None
Same = None
Max_M = None

def initializer(inter_adataI, sameI, thersholdI, max_M):
    global Inter_adata
    global Thershold
    global Same
    global Max_M
    Inter_adata = inter_adataI
    Thershold = thersholdI
    Same = sameI
    Max_M = 10000 if max_M < 2000 else max_M


def initializer_simple(inter_adataI):
    global Inter_adata
    Inter_adata = inter_adataI

def initializer_adata(adataI, sameI, thersholdI, max_M):
    global Adata
    global Thershold
    global Same
    global Max_M
    Adata = adataI
    Thershold = thersholdI
    Same = sameI
    Max_M = 10000 if max_M < 2000 else max_M




from setuptools import setup

setup(
    name='exoTras',
    version='0.4',
    author='Ruiqiao He',
    author_email='ruiqiaohe@gmail.com',
    packages=['exoTras'],
    license="GPL",
    url='http://pypi.python.org/pypi/exoTras/',
    description='exosome-containing droplet identification and source tracking in scRNA-seq data',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    install_requires=[
        "scanpy>=1.6.0",
        "numpy>=1.21.5",
        "pandas>=1.1.2",
        "scipy>=1.5.4",
        "statsmodels>=0.12.1",
        # "copy",
        # "sys",
        # "os",
        # "multiprocessing",
        # "pickle>=4.0",
        "gseapy",
    ],
    python_requires='>=3.7.1',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
      ],
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'exoTras=exoTras.main:exoTras_command',
        ]
    }
)



## source 
import gseapy as gp
import numpy as np
import pandas as pd
import scanpy as sc
import scipy
from scipy.spatial import cKDTree
import copy
import sys
import os
import pickle
from pathlib import Path

def source_biogenesis(adata_cell, OBScelltype='celltype', Xraw = True, normalW=True, species='Homo'):
    if Xraw:
        X_input = adata_cell.raw
    else:
        X_input = adata_cell
        
    if normalW:
        X_norm = sc.pp.scale(X_input.X, zero_center=True, max_value=None, copy=False)
    else:
        X_norm = X_input.X
    
    gsea_pval = []
    num_clusters = len(adata_cell.obs[OBScelltype].cat.categories)

    if species == 'Homo':
        gmt_path = Path(__file__).parent / 'exosomes.gmt'
    
    genesets_dict = {}
    with open(gmt_path) as genesets:
        genesets_dict = {
            line.strip().split("\t")[0]: line.strip().split("\t")[2:]
            for line in genesets.readlines()
        }

    for i in range(num_clusters):
        i = adata_cell.obs[OBScelltype].cat.categories[i]
        gene_rank = pd.DataFrame({'exp': np.array(X_norm[adata_cell.obs[OBScelltype] == str(i), :].mean(axis=0))}, index = X_input.var_names)

        res = gp.prerank(rnk=gene_rank, gene_sets=genesets_dict, outdir=None)
        terms = res.res2d.index
        gsea_pval.append([i, res.results[terms[0]]['nes'], res.results[terms[0]]['pval']])        

    gsea_pval_dat = pd.DataFrame(gsea_pval, index=adata_cell.obs[OBScelltype].cat.categories, columns = ['num', 'enrich', 'p'])
    gsea_pval_dat['log1p'] = -np.log10(gsea_pval_dat['p']+1e-4) * np.sign(gsea_pval_dat['enrich'])

    return(gsea_pval_dat)


def near_neighbor(adata_combined, OBSsample='batch', OBSexo='exo', OBScelltype='celltype', OBSMpca='X_pca', cellN=10):
    ## run sc.pp.pca(adata_combined) before
    near_neighbor = []
    for sample in adata_combined.obs[OBSsample].unique().astype(str):
        tse_ref = copy.copy(adata_combined[(adata_combined.obs[OBSexo] == '0') & (adata_combined.obs[OBSsample] == sample),])
        cell_tree = cKDTree(tse_ref.obsm[OBSMpca], leafsize=100)
        
        tmp_umap = adata_combined[(adata_combined.obs[OBSexo] == '1') & (adata_combined.obs[OBSsample] == sample),].obsm[OBSMpca]
        for i in range(tmp_umap.shape[0]):
            TheResult = cell_tree.query(tmp_umap[i,], k=10)

            near_neighbor.append([sample, i] + list(tse_ref.obs[OBScelltype].iloc[TheResult[1]]))#tmp_umap.obs['clusters'][i]] +

    near_neighbor_dat = pd.DataFrame(near_neighbor)
    near_neighbor_dat.columns = [str(i) for i in near_neighbor_dat.columns.values]
    return(near_neighbor_dat)

def preprocess_source(adata_exo, adata_cell, OBScelltype='celltype', OBSexo='exo', Xraw=True):
    ## cell type
    if Xraw:
        adata_cell_raw = copy.copy(adata_cell.raw.to_adata())
    else:
        adata_cell_raw = copy.copy(adata_cell)

    adata_exo.obs[OBScelltype] = 'Exosomes'
    adata_exo.obs[OBScelltype] = pd.Series(adata_exo.obs[OBScelltype], dtype="category")

    adata_combined = adata_cell_raw.concatenate(adata_exo, batch_key = OBSexo)

    adata_combined.obs[OBScelltype] = pd.Categorical(adata_combined.obs[OBScelltype], \
        categories = np.append(adata_cell_raw.obs[OBScelltype].cat.categories.values, 'Exosomes'), ordered = False)
    
    adata_combined.raw = adata_combined
    sc.pp.normalize_total(adata_combined, target_sum=1e4)
    sc.pp.log1p(adata_combined)
    sc.pp.highly_variable_genes(adata_combined, min_mean=0.0125, max_mean=3, min_disp=0.5)
    # sc.pl.highly_variable_genes(Normal_combined)
    adata_combined = adata_combined[:, adata_combined.var.highly_variable]#highly_variable

    sc.pp.pca(adata_combined)
    sc.pp.neighbors(adata_combined)
    sc.tl.umap(adata_combined)

    return(adata_combined)

def source_tracker(adata_exo, adata_cell, OBSsample='batch', OBScelltype='celltype', OBSexo='exo', OBSMpca='X_pca', cellN=10, Xraw = True, normalW=True):

    adata_combined = preprocess_source(adata_exo, adata_cell, OBScelltype=OBScelltype, OBSexo=OBSexo, Xraw=Xraw)
    gsea_pval_dat = source_biogenesis(adata_cell, OBScelltype=OBScelltype, Xraw = Xraw, normalW=normalW)
    near_neighbor_dat = near_neighbor(adata_combined, OBSsample=OBSsample, OBSexo=OBSexo, OBScelltype=OBScelltype, OBSMpca=OBSMpca, cellN=cellN)
    
    near_neighbor_dat['times'] = ''
    near_neighbor_dat['type'] = ''
    for i in range(near_neighbor_dat.shape[0]):
        ## iteration for all exosomes
        tmp_times = near_neighbor_dat.iloc[i,2:12].value_counts(sort = True)
        near_neighbor_dat.loc[i, 'times'] = tmp_times[0]
        tmp_keys = near_neighbor_dat.iloc[i,2:12].value_counts(sort = True).keys()
        tmp_times_fil = (tmp_times + gsea_pval_dat.loc[tmp_keys, 'log1p']*4).sort_values(ascending = False)
        near_neighbor_dat.loc[i, 'type'] = tmp_times_fil.keys()[0]

    near_neighbor_dat.index = adata_exo.obs.index
    celltype_e_number = pd.DataFrame(near_neighbor_dat.type.value_counts())

    adata_exo.obsm['source'] = near_neighbor_dat[['0', '1', 'type']]
    adata_exo.obsm['source'].columns = ['sample', 'i', 'type']

    return([celltype_e_number, adata_exo, adata_combined])


def ESAI_celltype(adata_exo, adata_cell, OBSsample='batch', OBScelltype='celltype'):
    ## calculate ESAI in cell type
    exo_type_count_total = pd.DataFrame(adata_exo.obsm['source'].groupby(['sample','type'])['sample'].count())
    cell_type_count = pd.DataFrame(adata_cell.obs.groupby([OBSsample, OBScelltype])[OBSsample].count())

    exo_activity = []
    for i, j in cell_type_count.index:
        if (i,j) in exo_type_count_total.index:
            a = exo_type_count_total.loc[(i, j), 'sample']
            b = cell_type_count.loc[(i, j), OBSsample]
            if b > 0:
                exo_activity.append([i, j, a, b, a / b])
            else:
                print([i, j, a, b, a / b])
        else:
            a = 0
            b = cell_type_count.loc[(i, j), OBSsample]
            exo_activity.append([i, j, a, b, 0])   
              
    exo_activity_dat = pd.DataFrame(exo_activity)
    exo_activity_dat_pivot = pd.pivot_table(exo_activity_dat, index=[0], columns= 1, values=4)

    return(exo_activity_dat_pivot)

